@@ -3,66 +3,65 @@
 
 from dateutil.relativedelta import relativedelta
 
+from trytond.cache import Cache
 from trytond.i18n import gettext
-from trytond.model import ModelView, ModelSQL, Workflow, fields
+from trytond.model import ModelSQL, ModelView, Workflow, fields
 from trytond.model.exceptions import AccessError
 from trytond.pool import Pool
-from trytond.pyson import Eval, If, PYSONEncoder, Id
+from trytond.pyson import Eval, Id
 from trytond.rpc import RPC
 from trytond.transaction import Transaction
 from trytond.wizard import (
-    Wizard, StateView, StateTransition, StateAction, Button)
+    Button, StateAction, StateTransition, StateView, Wizard)
 
-from .exceptions import (FiscalYearNotFoundError, FiscalYearDatesError,
-    FiscalYearSequenceError, FiscalYearCloseError, FiscalYearReOpenError)
+from .exceptions import (
+    FiscalYearCloseError, FiscalYearDatesError, FiscalYearNotFoundError,
+    FiscalYearReOpenError, FiscalYearSequenceError)
 
 STATES = {
     'readonly': Eval('state') != 'open',
 }
-DEPENDS = ['state']
 
 
 class FiscalYear(Workflow, ModelSQL, ModelView):
     'Fiscal Year'
     __name__ = 'account.fiscalyear'
-    name = fields.Char('Name', size=None, required=True, depends=DEPENDS)
+    name = fields.Char('Name', size=None, required=True)
     start_date = fields.Date('Starting Date', required=True, states=STATES,
-        domain=[('start_date', '<=', Eval('end_date', None))],
-        depends=DEPENDS + ['end_date'])
+        domain=[('start_date', '<=', Eval('end_date', None))])
     end_date = fields.Date('Ending Date', required=True, states=STATES,
-        domain=[('end_date', '>=', Eval('start_date', None))],
-        depends=DEPENDS + ['start_date'])
+        domain=[('end_date', '>=', Eval('start_date', None))])
     periods = fields.One2Many('account.period', 'fiscalyear', 'Periods',
         states=STATES,
         domain=[
             ('company', '=', Eval('company')),
             ],
-        depends=DEPENDS + ['company'])
+        order=[('start_date', 'ASC'), ('id', 'ASC')])
     state = fields.Selection([
             ('open', 'Open'),
-            ('close', 'Close'),
+            ('closed', 'Closed'),
             ('locked', 'Locked'),
-            ], 'State', readonly=True, required=True)
+            ], 'State', readonly=True, required=True, sort=False)
     post_move_sequence = fields.Many2One(
         'ir.sequence', "Post Move Sequence", required=True,
         domain=[
             ('sequence_type', '=',
                 Id('account', 'sequence_type_account_move')),
             ('company', '=', Eval('company')),
-            ],
-        depends=['company'])
+            ])
     company = fields.Many2One(
-        'company.company', "Company", required=True, select=True)
+        'company.company', "Company", required=True)
     icon = fields.Function(fields.Char("Icon"), 'get_icon')
+    _find_cache = Cache(__name__ + '.find', context=False)
 
     @classmethod
     def __setup__(cls):
         super(FiscalYear, cls).__setup__()
-        cls._order.insert(0, ('start_date', 'ASC'))
+        cls._order.insert(0, ('start_date', 'DESC'))
         cls._transitions |= set((
-                ('open', 'close'),
-                ('close', 'locked'),
-                ('close', 'open'),
+                ('open', 'closed'),
+                ('closed', 'locked'),
+                ('closed', 'open'),
                 ))
         cls._buttons.update({
                 'create_periods': {
@@ -75,11 +74,11 @@ class FiscalYear(Workflow, ModelSQL, ModelView):
                     'depends': ['state'],
                     },
                 'reopen': {
-                    'invisible': Eval('state') != 'close',
+                    'invisible': Eval('state') != 'closed',
                     'depends': ['state'],
                     },
-                'lock': {
-                    'invisible': Eval('state') != 'close',
+                'lock_': {
+                    'invisible': Eval('state') != 'closed',
                     'depends': ['state'],
                     },
                 })
@@ -87,6 +86,15 @@ class FiscalYear(Workflow, ModelSQL, ModelView):
                 'create_period': RPC(readonly=False, instantiate=0),
                 })
 
+    @classmethod
+    def __register__(cls, module):
+        cursor = Transaction().connection.cursor()
+        t = cls.__table__()
+        super().__register__(module)
+        # Migration from 6.8: rename state close to closed
+        cursor.execute(
+            *t.update([t.state], ['closed'], where=t.state == 'close'))
+
     @staticmethod
     def default_state():
         return 'open'
@@ -98,50 +106,79 @@ class FiscalYear(Workflow, ModelSQL, ModelView):
     def get_icon(self, name):
         return {
             'open': 'tryton-account-open',
-            'close': 'tryton-account-close',
+            'closed': 'tryton-account-close',
             'locked': 'tryton-account-block',
             }.get(self.state)
 
     @classmethod
-    def validate(cls, years):
-        super(FiscalYear, cls).validate(years)
-        for year in years:
-            year.check_dates()
-            year.check_post_move_sequence()
+    def validate_fields(cls, fiscalyears, field_names):
+        super().validate_fields(fiscalyears, field_names)
+        cls.check_dates(fiscalyears, field_names)
+        cls.check_post_move_sequence(fiscalyears, field_names)
 
-    def check_dates(self):
+    @classmethod
+    def check_dates(cls, fiscalyears, field_names=None):
+        if field_names and not (field_names & {
+                    'start_date', 'end_date', 'state', 'company'}):
+            return
         transaction = Transaction()
         connection = transaction.connection
-        transaction.database.lock(connection, self._table)
+        cls.lock()
         cursor = connection.cursor()
-        table = self.__table__()
-        cursor.execute(*table.select(table.id,
-                where=(((table.start_date <= self.start_date)
-                        & (table.end_date >= self.start_date))
-                    | ((table.start_date <= self.end_date)
-                        & (table.end_date >= self.end_date))
-                    | ((table.start_date >= self.start_date)
-                        & (table.end_date <= self.end_date)))
-                & (table.company == self.company.id)
-                & (table.id != self.id)))
-        second_id = cursor.fetchone()
-        if second_id:
-            second = self.__class__(second_id[0])
-            raise FiscalYearDatesError(
-                gettext('account.msg_fiscalyear_overlap',
-                    first=self.rec_name,
-                    second=second.rec_name))
+        table = cls.__table__()
+        for year in fiscalyears:
+            cursor.execute(*table.select(table.id,
+                    where=(((table.start_date <= year.start_date)
+                            & (table.end_date >= year.start_date))
+                        | ((table.start_date <= year.end_date)
+                            & (table.end_date >= year.end_date))
+                        | ((table.start_date >= year.start_date)
+                            & (table.end_date <= year.end_date)))
+                    & (table.company == year.company.id)
+                    & (table.id != year.id)))
+            second_id = cursor.fetchone()
+            if second_id:
+                second = cls(second_id[0])
+                raise FiscalYearDatesError(
+                    gettext('account.msg_fiscalyear_overlap',
+                        first=year.rec_name,
+                        second=second.rec_name))
+            if year.state == 'open':
+                fiscalyears = cls.search([
+                        ('start_date', '>=', year.end_date),
+                        ('state', 'in', ['closed', 'locked']),
+                        ('company', '=', year.company.id),
+                        ],
+                    limit=1,
+                    order=[('start_date', 'ASC')])
+                if fiscalyears:
+                    fiscalyear, = fiscalyears
+                    raise FiscalYearDatesError(
+                        gettext('account.msg_open_fiscalyear_earlier',
+                            open=year.rec_name,
+                            closed=fiscalyear.rec_name))
 
-    def check_post_move_sequence(self):
-        years = self.search([
-                ('post_move_sequence', '=', self.post_move_sequence.id),
-                ('id', '!=', self.id),
-                ])
-        if years:
-            raise FiscalYearSequenceError(
-                gettext('account.msg_fiscalyear_different_post_move_sequence',
-                    first=self.rec_name,
-                    second=years[0].rec_name))
+    @classmethod
+    def check_post_move_sequence(cls, fiscalyears, field_names=None):
+        if field_names and 'post_move_sequence' not in field_names:
+            return
+        for fiscalyear in fiscalyears:
+            sequence = fiscalyear.post_move_sequence
+            years = cls.search([
+                    ('post_move_sequence', '=', sequence.id),
+                    ('id', '!=', fiscalyear.id),
+                    ], limit=1)
+            if years:
+                raise FiscalYearSequenceError(gettext(
+                        'account.msg_fiscalyear_different_post_move_sequence',
+                        first=fiscalyear.rec_name,
+                        second=years[0].rec_name))
+
+    @classmethod
+    def create(cls, vlist):
+        fiscalyears = super().create(vlist)
+        cls._find_cache.clear()
+        return fiscalyears
 
     @classmethod
     def write(cls, *args):
@@ -163,12 +200,14 @@ class FiscalYear(Workflow, ModelSQL, ModelView):
                                     'msg_change_fiscalyear_post_move_sequence',
                                     fiscalyear=fiscalyear.rec_name))
         super(FiscalYear, cls).write(*args)
+        cls._find_cache.clear()
 
     @classmethod
     def delete(cls, fiscalyears):
         Period = Pool().get('account.period')
         Period.delete([p for f in fiscalyears for p in f.periods])
         super(FiscalYear, cls).delete(fiscalyears)
+        cls._find_cache.clear()
 
     @classmethod
     def create_period(cls, fiscalyears, interval=1, end_day=31):
@@ -206,33 +245,56 @@ class FiscalYear(Workflow, ModelSQL, ModelView):
         pass
 
     @classmethod
-    def find(cls, company_id, date=None, exception=True):
+    def find(cls, company, date=None, test_state=True):
         '''
-        Return the fiscal year for the company_id
-            at the date or the current date.
-        If exception is set the function will raise an exception
-            if any fiscal year is found.
+        Return the fiscal year for the company at the date or the current date
+        or raise FiscalYearNotFoundError.
+        If test_state is true, it searches on non-closed fiscal years
         '''
         pool = Pool()
         Lang = pool.get('ir.lang')
         Date = pool.get('ir.date')
+        Company = pool.get('company.company')
 
+        company_id = int(company) if company is not None else None
         if not date:
-            date = Date.today()
-        fiscalyears = cls.search([
-            ('start_date', '<=', date),
-            ('end_date', '>=', date),
-            ('company', '=', company_id),
-            ], order=[('start_date', 'DESC')], limit=1)
-        if not fiscalyears:
-            if exception:
-                lang = Lang.get()
+            with Transaction().set_context(company=company_id):
+                date = Date.today()
+        key = (company_id, date)
+        fiscalyear = cls._find_cache.get(key, -1)
+        if fiscalyear is not None and fiscalyear < 0:
+            clause = [
+                ('start_date', '<=', date),
+                ('end_date', '>=', date),
+                ('company', '=', company_id),
+                ]
+            fiscalyears = cls.search(
+                clause, order=[('start_date', 'DESC')], limit=1)
+            if fiscalyears:
+                fiscalyear, = fiscalyears
+            else:
+                fiscalyear = None
+            cls._find_cache.set(key, int(fiscalyear) if fiscalyear else None)
+        elif fiscalyear is not None:
+            fiscalyear = cls(fiscalyear)
+        found = fiscalyear and (not test_state or fiscalyear.state == 'open')
+        if not found:
+            lang = Lang.get()
+            if company is not None and not isinstance(company, Company):
+                company = Company(company)
+            if not fiscalyear:
                 raise FiscalYearNotFoundError(
                     gettext('account.msg_no_fiscalyear_date',
-                        date=lang.strftime(date)))
+                        date=lang.strftime(date),
+                        company=company.rec_name if company else ''))
             else:
-                return None
-        return fiscalyears[0].id
+                raise FiscalYearNotFoundError(
+                    gettext('account.msg_no_open_fiscalyear_date',
+                        date=lang.strftime(date),
+                        fiscalyear=fiscalyear.rec_name,
+                        company=company.rec_name if company else ''))
+        else:
+            return fiscalyear
 
     def get_deferral(self, account):
         'Computes deferrals for accounts'
@@ -254,12 +316,13 @@ class FiscalYear(Workflow, ModelSQL, ModelView):
             deferral.fiscalyear = self
             deferral.debit = account.debit
             deferral.credit = account.credit
+            deferral.line_count = account.line_count
             deferral.amount_second_currency = account.amount_second_currency
             return deferral
 
     @classmethod
     @ModelView.button
-    @Workflow.transition('close')
+    @Workflow.transition('closed')
     def close(cls, fiscalyears):
         '''
         Close a fiscal year
@@ -268,12 +331,10 @@ class FiscalYear(Workflow, ModelSQL, ModelView):
         Period = pool.get('account.period')
         Account = pool.get('account.account')
         Deferral = pool.get('account.account.deferral')
-        transaction = Transaction()
-        database = transaction.database
-        connection = transaction.connection
 
-        # Lock period to be sure no new period will be created in between.
-        database.lock(connection, Period._table)
+        # Prevent create new fiscal year or period
+        cls.lock()
+        Period.lock()
 
         deferrals = []
         for fiscalyear in fiscalyears:
@@ -296,8 +357,10 @@ class FiscalYear(Workflow, ModelSQL, ModelView):
                 accounts = Account.search([
                         ('company', '=', fiscalyear.company.id),
                         ])
-                deferrals += [_f for _f in (fiscalyear.get_deferral(a)
-                        for a in accounts) if _f]
+                for account in accounts:
+                    deferral = fiscalyear.get_deferral(account)
+                    if deferral:
+                        deferrals.append(deferral)
         Deferral.save(deferrals)
 
     @classmethod
@@ -327,13 +390,13 @@ class FiscalYear(Workflow, ModelSQL, ModelView):
     @classmethod
     @ModelView.button
     @Workflow.transition('locked')
-    def lock(cls, fiscalyears):
+    def lock_(cls, fiscalyears):
         pool = Pool()
         Period = pool.get('account.period')
         periods = Period.search([
                 ('fiscalyear', 'in', [f.id for f in fiscalyears]),
                 ])
-        Period.lock(periods)
+        Period.lock_(periods)
 
 
 class BalanceNonDeferralStart(ModelView):
@@ -350,13 +413,12 @@ class BalanceNonDeferralStart(ModelView):
         context={
             'company': Eval('company', -1),
             },
-        depends=['company'])
+        depends={'company'})
     period = fields.Many2One('account.period', 'Period', required=True,
         domain=[
             ('fiscalyear', '=', Eval('fiscalyear')),
             ('type', '=', 'adjustment'),
-            ],
-        depends=['fiscalyear'])
+            ])
     credit_account = fields.Many2One('account.account', 'Credit Account',
         required=True,
         domain=[
@@ -364,8 +426,7 @@ class BalanceNonDeferralStart(ModelView):
             ('closed', '!=', True),
             ('company', '=', Eval('company', -1)),
             ('deferral', '=', True),
-            ],
-        depends=['company'])
+            ])
     debit_account = fields.Many2One('account.account', 'Debit Account',
         required=True,
         domain=[
@@ -373,13 +434,11 @@ class BalanceNonDeferralStart(ModelView):
             ('closed', '!=', True),
             ('company', '=', Eval('company', -1)),
             ('deferral', '=', True),
-            ],
-        depends=['company'])
+            ])
 
     @fields.depends('fiscalyear')
     def on_change_with_company(self, name=None):
-        if self.fiscalyear:
-            return self.fiscalyear.company.id
+        return self.fiscalyear.company if self.fiscalyear else None
 
 
 class BalanceNonDeferral(Wizard):
@@ -390,7 +449,7 @@ class BalanceNonDeferral(Wizard):
             Button('Cancel', 'end', 'tryton-cancel'),
             Button('OK', 'balance', 'tryton-ok', default=True),
             ])
-    balance = StateAction('account.act_move_line_form')
+    balance = StateAction('account.act_move_form')
 
     def get_move_line(self, account):
         pool = Pool()
@@ -460,11 +519,10 @@ class BalanceNonDeferral(Wizard):
         return move
 
     def do_balance(self, action):
-        self.create_move()
-        action['pyson_domain'] = PYSONEncoder().encode([
-                ('move.origin', '=', str(self.start.fiscalyear)),
-                ])
-        return action, {}
+        move = self.create_move()
+        if move:
+            action['views'].reverse()
+        return action, {'res_id': move.id if move else None}
 
 
 class CreatePeriodsStart(ModelView):
@@ -479,7 +537,6 @@ class CreatePeriodsStart(ModelView):
         states={
             'invisible': Eval('frequency') != 'other',
             },
-        depends=['frequency'],
         help="The length of each period, in months.")
     end_day = fields.Integer("End Day", required=True,
         help="The day of the month on which periods end.\n"
@@ -538,7 +595,6 @@ class RenewFiscalYearStart(ModelView):
         domain=[
             ('company', '=', Eval('company')),
             ],
-        depends=['company'],
         help="Used as reference for fiscalyear configuration.")
     start_date = fields.Date("Start Date", required=True)
     end_date = fields.Date("End Date", required=True)
@@ -635,6 +691,6 @@ class RenewFiscalYear(Wizard):
 
     def do_create_(self, action):
         fiscalyear = self.create_fiscalyear()
-        action['res_id'] = [fiscalyear.id]
+        fiscalyear.save()
         action['views'].reverse()
-        return action, {}
+        return action, {'res_id': fiscalyear.id}
