@@ -3,25 +3,27 @@
 import datetime
 from collections import namedtuple
 from decimal import Decimal
-from itertools import groupby
+from itertools import cycle, groupby
 
 from sql import Literal
 from sql.aggregate import Sum
 from sql.conditionals import Case
 
+from trytond import backend
 from trytond.i18n import gettext
 from trytond.model import (
-    ModelView, ModelSQL, MatchMixin, DeactivableMixin, fields,
+    DeactivableMixin, Index, MatchMixin, ModelSQL, ModelView, fields,
     sequence_ordered, tree)
 from trytond.model.exceptions import AccessError
-from trytond.wizard import Wizard, StateView, StateAction, Button
-from trytond import backend
-from trytond.pyson import Eval, If, Bool, PYSONEncoder
-from trytond.transaction import Transaction
-from trytond.tools import cursor_dict, lstrip_wildcard
+from trytond.modules.currency.fields import Monetary
 from trytond.pool import Pool
+from trytond.pyson import Bool, Eval, If, PYSONEncoder
+from trytond.tools import cursor_dict, is_full_text, lstrip_wildcard
+from trytond.transaction import Transaction
+from trytond.wizard import Button, StateAction, StateView, Wizard
 
-from .common import PeriodMixin, ActivePeriodMixin
+from .common import ActivePeriodMixin, ContextCompanyMixin, PeriodMixin
+from .exceptions import PeriodNotFoundError
 
 KINDS = [
     ('sale', 'Sale'),
@@ -120,42 +122,49 @@ class TaxCodeTemplate(PeriodMixin, tree(), ModelSQL, ModelView):
             childs = sum((c.childs for c in childs), ())
 
 
-class TaxCode(ActivePeriodMixin, tree(), ModelSQL, ModelView):
+class TaxCode(
+        ContextCompanyMixin, ActivePeriodMixin, tree(), ModelSQL, ModelView):
     'Tax Code'
     __name__ = 'account.tax.code'
     _states = {
         'readonly': (Bool(Eval('template', -1))
             & ~Eval('template_override', False)),
         }
-    name = fields.Char('Name', required=True, select=True, states=_states)
-    code = fields.Char('Code', select=True, states=_states)
-    company = fields.Many2One('company.company', 'Company', required=True,
-        select=True)
+    name = fields.Char('Name', required=True, states=_states)
+    code = fields.Char('Code', states=_states)
+    company = fields.Many2One('company.company', 'Company', required=True)
     parent = fields.Many2One(
-        'account.tax.code', 'Parent', select=True, states=_states)
+        'account.tax.code', "Parent", states=_states,
+        domain=[
+            ('company', '=', Eval('company', -1)),
+            ])
     lines = fields.One2Many('account.tax.code.line', 'code', "Lines")
-    childs = fields.One2Many('account.tax.code', 'parent', 'Children')
+    childs = fields.One2Many(
+        'account.tax.code', 'parent', "Children",
+        domain=[
+            ('company', '=', Eval('company', -1)),
+            ])
     currency = fields.Function(fields.Many2One('currency.currency',
         'Currency'), 'on_change_with_currency')
-    currency_digits = fields.Function(fields.Integer('Currency Digits'),
-        'on_change_with_currency_digits')
-    amount = fields.Function(fields.Numeric(
-            "Amount", digits=(16, Eval('currency_digits', 2)),
-            depends=['currency_digits']),
+    amount = fields.Function(Monetary(
+            "Amount", currency='currency', digits='currency'),
         'get_amount')
     template = fields.Many2One('account.tax.code.template', 'Template')
     template_override = fields.Boolean('Override Template',
         help="Check to override template definition",
         states={
             'invisible': ~Bool(Eval('template', -1)),
-            },
-        depends=['template'])
+            })
     description = fields.Text('Description')
     del _states
 
     @classmethod
     def __setup__(cls):
+        cls.code.search_unaccented = False
         super(TaxCode, cls).__setup__()
+        t = cls.__table__()
+        cls._sql_indexes.add(
+            Index(t, (t.code, Index.Similarity())))
         for date in [cls.start_date, cls.end_date]:
             date.states = {
                 'readonly': (Bool(Eval('template', -1))
@@ -173,14 +182,7 @@ class TaxCode(ActivePeriodMixin, tree(), ModelSQL, ModelView):
 
     @fields.depends('company')
     def on_change_with_currency(self, name=None):
-        if self.company:
-            return self.company.currency.id
-
-    @fields.depends('currency')
-    def on_change_with_currency_digits(self, name=None):
-        if self.currency:
-            return self.currency.digits
-        return 2
+        return self.company.currency if self.company else None
 
     @classmethod
     def get_amount(cls, codes, name):
@@ -191,9 +193,8 @@ class TaxCode(ActivePeriodMixin, tree(), ModelSQL, ModelView):
                 ('parent', 'child_of', [c.id for c in codes]),
                 ])
         for code in childs:
-            exp = Decimal(str(10.0 ** -code.currency_digits))
-            result[code.id] = sum(
-                (l.value for l in code.lines), Decimal(0)).quantize(exp)
+            result[code.id] = code.currency.round(
+                sum((l.value for l in code.lines), Decimal(0)))
             parents[code.id] = code.parent.id if code.parent else None
 
         ids = set(map(int, childs))
@@ -222,16 +223,17 @@ class TaxCode(ActivePeriodMixin, tree(), ModelSQL, ModelView):
 
     @classmethod
     def search_rec_name(cls, name, clause):
-        if clause[1].startswith('!') or clause[1].startswith('not '):
+        _, operator, operand, *extra = clause
+        if operator.startswith('!') or operator.startswith('not '):
             bool_op = 'AND'
         else:
             bool_op = 'OR'
-        code_value = clause[2]
-        if clause[1].endswith('like'):
-            code_value = lstrip_wildcard(clause[2])
+        code_value = operand
+        if operator.endswith('like') and is_full_text(operand):
+            code_value = lstrip_wildcard(operand)
         return [bool_op,
-            ('code', clause[1], code_value) + tuple(clause[3:]),
-            (cls._rec_name,) + tuple(clause[1:]),
+            ('code', operator, code_value, *extra),
+            (cls._rec_name, operator, operand, *extra),
             ]
 
     @classmethod
@@ -378,7 +380,8 @@ class TaxCodeLine(ModelSQL, ModelView):
         'account.tax', "Tax", required=True, states=_states,
         domain=[
             ('company', '=', Eval('_parent_code', {}).get('company')),
-            ])
+            ],
+        depends={'code'})
     amount = fields.Selection([
             ('tax', "Tax"),
             ('base', "Base"),
@@ -394,8 +397,7 @@ class TaxCodeLine(ModelSQL, ModelView):
         help="Check to override template definition",
         states={
             'invisible': ~Bool(Eval('template', -1)),
-            },
-        depends=['template'])
+            })
 
     del _states
 
@@ -472,61 +474,117 @@ class TaxCodeLine(ModelSQL, ModelView):
             cls.write(*values)
 
 
-class OpenChartTaxCodeStart(ModelView):
-    'Open Chart of Tax Codes'
-    __name__ = 'account.tax.code.open_chart.start'
+class TaxCodeContext(ModelView):
+    "Tax Code Context"
+    __name__ = 'account.tax.code.context'
+
+    company = fields.Many2One('company.company', "Company", required=True)
     method = fields.Selection([
-        ('fiscalyear', 'By Fiscal Year'),
-        ('periods', 'By Periods'),
-        ], 'Method', required=True)
-    fiscalyear = fields.Many2One('account.fiscalyear', 'Fiscal Year',
-        help='Leave empty for all open fiscal year.',
+            ('fiscalyear', "By Fiscal Year"),
+            ('period', "By Period"),
+            ('periods', "Over Periods"),
+            ], 'Method', required=True)
+
+    fiscalyear = fields.Many2One(
+        'account.fiscalyear', "Fiscal Year",
+        domain=[
+            ('company', '=', Eval('company', -1)),
+            ],
         states={
             'invisible': Eval('method') != 'fiscalyear',
             'required': Eval('method') == 'fiscalyear',
-            },
-        depends=['method'])
-    periods = fields.Many2Many('account.period', None, None, 'Periods',
-        help='Leave empty for all periods of all open fiscal year.',
+            })
+
+    period = fields.Many2One(
+        'account.period', "Period",
+        domain=[
+            ('company', '=', Eval('company', -1)),
+            ('type', '=', 'standard'),
+            ],
+        states={
+            'invisible': Eval('method') != 'period',
+            'required': Eval('method') == 'period',
+            })
+
+    start_period = fields.Many2One(
+        'account.period', "Start Period",
+        domain=[
+            ('company', '=', Eval('company', -1)),
+            ('type', '=', 'standard'),
+            ('start_date', '<=', (Eval('end_period'), 'start_date')),
+            ],
         states={
             'invisible': Eval('method') != 'periods',
             'required': Eval('method') == 'periods',
-            },
-        depends=['method'])
+            })
+    end_period = fields.Many2One(
+        'account.period', "End Period",
+        domain=[
+            ('company', '=', Eval('company', -1)),
+            ('type', '=', 'standard'),
+            ('start_date', '>=', (Eval('start_period'), 'start_date')),
+            ],
+        states={
+            'invisible': Eval('method') != 'periods',
+            'required': Eval('method') == 'periods',
+            })
 
-    @staticmethod
-    def default_method():
-        return 'periods'
-
-
-class OpenChartTaxCode(Wizard):
-    'Open Chart of Tax Codes'
-    __name__ = 'account.tax.code.open_chart'
-    start = StateView('account.tax.code.open_chart.start',
-        'account.tax_code_open_chart_start_view_form', [
-            Button('Cancel', 'end', 'tryton-cancel'),
-            Button('Open', 'open_', 'tryton-ok', default=True),
+    periods = fields.Many2Many(
+        'account.period', None, None, "Periods",
+        domain=[
+            ('company', '=', Eval('company', -1)),
+            ('type', '=', 'standard'),
             ])
-    open_ = StateAction('account.act_tax_code_tree2')
 
-    def do_open_(self, action):
+    @classmethod
+    def default_company(cls):
+        return Transaction().context.get('company')
+
+    @fields.depends(
+        'company', 'fiscalyear', 'period', methods=['on_change_with_periods'])
+    def on_change_company(self):
+        if self.fiscalyear and self.fiscalyear.company != self.company:
+            self.fiscalyear = None
+        if self.period and self.period.company != self.company:
+            self.period = None
+        self.periods = self.on_change_with_periods()
+
+    @classmethod
+    def default_method(cls):
+        return 'period'
+
+    @classmethod
+    def default_period(cls):
+        pool = Pool()
+        Period = pool.get('account.period')
+        try:
+            period = Period.find(cls.default_company(), test_state=False)
+        except PeriodNotFoundError:
+            return None
+        return period.id
+
+    @fields.depends(
+        'method', 'company',
+        'fiscalyear', 'period', 'start_period', 'end_period')
+    def on_change_with_periods(self):
+        pool = Pool()
+        Period = pool.get('account.period')
         periods = []
-        if self.start.method == 'fiscalyear':
-            if self.start.fiscalyear:
-                periods = self.start.fiscalyear.periods
-                action['name'] += ' - %s' % self.start.fiscalyear.rec_name
-        elif self.start.method == 'periods':
-            if self.start.periods:
-                periods = self.start.periods
-                action['name'] += ' (%s)' % ', '.join(
-                    p.rec_name for p in self.start.periods)
-        action['pyson_context'] = PYSONEncoder().encode({
-                'periods': [p.id for p in periods],
-                })
-        return action, {}
-
-    def transition_open_(self):
-        return 'end'
+        if self.method == 'fiscalyear' and self.fiscalyear:
+            periods.extend(
+                p for p in self.fiscalyear.periods if p.type == 'standard')
+        elif self.method == 'period' and self.period:
+            periods.append(self.period)
+        elif (self.method == 'periods'
+                and self.company
+                and self.start_period and self.end_period):
+            periods = Period.search([
+                    ('start_date', '>=', self.start_period.start_date),
+                    ('start_date', '<=', self.end_period.start_date),
+                    ('company', '=', self.company.id),
+                    ('type', '=', 'standard'),
+                    ])
+        return periods
 
 
 class TaxTemplate(sequence_ordered(), ModelSQL, ModelView, DeactivableMixin):
@@ -538,21 +596,19 @@ class TaxTemplate(sequence_ordered(), ModelSQL, ModelView, DeactivableMixin):
         'account.tax.group', 'Group',
         states={
             'invisible': Bool(Eval('parent')),
-            },
-        depends=['parent'])
+            })
     start_date = fields.Date('Starting Date')
     end_date = fields.Date('Ending Date')
     amount = fields.Numeric('Amount', digits=(16, 8),
         states={
             'required': Eval('type') == 'fixed',
             'invisible': Eval('type') != 'fixed',
-            },
-        depends=['type'])
+            })
     rate = fields.Numeric('Rate', digits=(14, 10),
         states={
             'required': Eval('type') == 'percentage',
             'invisible': Eval('type') != 'percentage',
-            }, depends=['type'])
+            })
     type = fields.Selection([
         ('percentage', 'Percentage'),
         ('fixed', 'Fixed'),
@@ -561,8 +617,7 @@ class TaxTemplate(sequence_ordered(), ModelSQL, ModelView, DeactivableMixin):
     update_unit_price = fields.Boolean('Update Unit Price',
         states={
             'invisible': Bool(Eval('parent')),
-            },
-        depends=['parent'])
+            })
     parent = fields.Many2One('account.tax.template', 'Parent')
     childs = fields.One2Many('account.tax.template', 'parent', 'Children')
     invoice_account = fields.Many2One(
@@ -572,10 +627,8 @@ class TaxTemplate(sequence_ordered(), ModelSQL, ModelView, DeactivableMixin):
             ('closed', '!=', True),
             ],
         states={
-            'invisible': Eval('type') == 'none',
             'required': Eval('type') != 'none',
-            },
-        depends=['type'])
+            })
     credit_note_account = fields.Many2One(
         'account.account.template', 'Credit Note Account',
         domain=[
@@ -583,10 +636,8 @@ class TaxTemplate(sequence_ordered(), ModelSQL, ModelView, DeactivableMixin):
             ('closed', '!=', True),
             ],
         states={
-            'invisible': Eval('type') == 'none',
             'required': Eval('type') != 'none',
-            },
-        depends=['type'])
+            })
     account = fields.Many2One('account.account.template', 'Account Template',
             domain=[('parent', '=', None)], required=True)
     legal_notice = fields.Text("Legal Notice")
@@ -597,16 +648,19 @@ class TaxTemplate(sequence_ordered(), ModelSQL, ModelView, DeactivableMixin):
         cls._order.insert(0, ('account', 'ASC'))
 
     @classmethod
-    def validate(cls, tax_templates):
-        super(TaxTemplate, cls).validate(tax_templates)
-        for tax_template in tax_templates:
-            tax_template.check_update_unit_price()
+    def validate_fields(cls, tax_templates, field_names):
+        super().validate_fields(tax_templates, field_names)
+        cls.check_update_unit_price(tax_templates, field_names)
 
-    def check_update_unit_price(self):
-        if self.update_unit_price and self.parent:
-            raise AccessError(
-                gettext('account.msg_tax_update_unit_price_with_parent',
-                    tax=self.rec_name))
+    @classmethod
+    def check_update_unit_price(cls, tax_templates, field_names=None):
+        if field_names and not (field_names & {'update_unit_price', 'parent'}):
+            return
+        for tax in tax_templates:
+            if tax.update_unit_price and tax.parent:
+                raise AccessError(gettext(
+                        'account.msg_tax_update_unit_price_with_parent',
+                        tax=tax.rec_name))
 
     @staticmethod
     def default_type():
@@ -714,7 +768,7 @@ class Tax(sequence_ordered(), ModelSQL, ModelView, DeactivableMixin):
         states={
             'invisible': Bool(Eval('parent')),
             'readonly': _states['readonly'],
-            }, depends=['parent'])
+            })
     start_date = fields.Date('Starting Date', states=_states)
     end_date = fields.Date('Ending Date', states=_states)
     amount = fields.Numeric('Amount', digits=(16, 8),
@@ -722,14 +776,13 @@ class Tax(sequence_ordered(), ModelSQL, ModelView, DeactivableMixin):
             'required': Eval('type') == 'fixed',
             'invisible': Eval('type') != 'fixed',
             'readonly': _states['readonly'],
-            }, help='In company\'s currency.',
-        depends=['type'])
+            }, help='In company\'s currency.')
     rate = fields.Numeric('Rate', digits=(14, 10),
         states={
             'required': Eval('type') == 'percentage',
             'invisible': Eval('type') != 'percentage',
             'readonly': _states['readonly'],
-            }, depends=['type'])
+            })
     type = fields.Selection([
         ('percentage', 'Percentage'),
         ('fixed', 'Fixed'),
@@ -740,14 +793,19 @@ class Tax(sequence_ordered(), ModelSQL, ModelView, DeactivableMixin):
             'invisible': Bool(Eval('parent')),
             'readonly': _states['readonly'],
             },
-        depends=['parent'],
         help=('If checked then the unit price for further tax computation will'
             ' be modified by this tax.'))
-    parent = fields.Many2One('account.tax', 'Parent', ondelete='CASCADE',
-        states=_states)
-    childs = fields.One2Many('account.tax', 'parent', 'Children')
-    company = fields.Many2One(
-        'company.company', "Company", required=True, select=True)
+    parent = fields.Many2One(
+        'account.tax', "Parent", ondelete='CASCADE', states=_states,
+        domain=[
+            ('company', '=', Eval('company', -1)),
+            ])
+    childs = fields.One2Many(
+        'account.tax', 'parent', "Children",
+        domain=[
+            ('company', '=', Eval('company', -1)),
+            ])
+    company = fields.Many2One('company.company', "Company", required=True)
     invoice_account = fields.Many2One('account.account', 'Invoice Account',
         domain=[
             ('company', '=', Eval('company')),
@@ -756,10 +814,8 @@ class Tax(sequence_ordered(), ModelSQL, ModelView, DeactivableMixin):
             ],
         states={
             'readonly': _states['readonly'],
-            'invisible': Eval('type') == 'none',
-            'required': (Eval('type') != 'none') & Eval('company'),
-            },
-        depends=['company', 'type'])
+            'required': Eval('type') != 'none',
+            })
     credit_note_account = fields.Many2One('account.account',
         'Credit Note Account',
         domain=[
@@ -769,10 +825,8 @@ class Tax(sequence_ordered(), ModelSQL, ModelView, DeactivableMixin):
             ],
         states={
             'readonly': _states['readonly'],
-            'invisible': Eval('type') == 'none',
-            'required': (Eval('type') != 'none') & Eval('company'),
-            },
-        depends=['company', 'type'])
+            'required': Eval('type') != 'none',
+            })
     legal_notice = fields.Text("Legal Notice", translate=True,
         states=_states)
     template = fields.Many2One('account.tax.template', 'Template',
@@ -781,42 +835,41 @@ class Tax(sequence_ordered(), ModelSQL, ModelView, DeactivableMixin):
         help="Check to override template definition",
         states={
             'invisible': ~Bool(Eval('template', -1)),
-            },
-        depends=['template'])
+            })
 
-    invoice_base_amount = fields.Function(fields.Numeric(
-            "Invoice Base Amount", digits=(16, Eval('currency_digits', 2)),
-            depends=['currency_digits']),
+    invoice_base_amount = fields.Function(Monetary(
+            "Invoice Base Amount", currency='currency', digits='currency'),
         'get_amount')
-    invoice_tax_amount = fields.Function(fields.Numeric(
-            "Invoice Tax Amount", digits=(16, Eval('currency_digits', 2)),
-            depends=['currency_digits']),
+    invoice_tax_amount = fields.Function(Monetary(
+            "Invoice Tax Amount", currency='currency', digits='currency'),
         'get_amount')
-    credit_base_amount = fields.Function(fields.Numeric(
-            "Credit Base Amount", digits=(16, Eval('currency_digits', 2)),
-            depends=['currency_digits']),
+    credit_base_amount = fields.Function(Monetary(
+            "Credit Base Amount", currency='currency', digits='currency'),
         'get_amount')
-    credit_tax_amount = fields.Function(fields.Numeric(
-            "Credit Tax Amount", digits=(16, Eval('currency_digits', 2)),
-            depends=['currency_digits']),
+    credit_tax_amount = fields.Function(Monetary(
+            "Credit Tax Amount", currency='currency', digits='currency'),
         'get_amount')
 
-    currency_digits = fields.Function(fields.Integer("Currency Digits"),
-        'on_change_with_currency_digits')
+    currency = fields.Function(fields.Many2One(
+            'currency.currency', "Currency"),
+        'on_change_with_currency')
 
     del _states
 
     @classmethod
-    def validate(cls, taxes):
-        super(Tax, cls).validate(taxes)
-        for tax in taxes:
-            tax.check_update_unit_price()
+    def validate_fields(cls, taxes, field_names):
+        super().validate_fields(taxes, field_names)
+        cls.check_update_unit_price(taxes, field_names)
 
-    def check_update_unit_price(self):
-        if self.parent and self.update_unit_price:
-            raise AccessError(
-                gettext('account.msg_tax_update_unit_price_with_parent',
-                    tax=self.rec_name))
+    @classmethod
+    def check_update_unit_price(cls, taxes, field_names=None):
+        if field_names and not (field_names & {'parent', 'update_unit_price'}):
+            return
+        for tax in taxes:
+            if tax.parent and tax.update_unit_price:
+                raise AccessError(gettext(
+                        'account.msg_tax_update_unit_price_with_parent',
+                        tax=tax.rec_name))
 
     @staticmethod
     def default_type():
@@ -921,10 +974,8 @@ class Tax(sequence_ordered(), ModelSQL, ModelView, DeactivableMixin):
         return [('move_line.move.period', 'in', periods)]
 
     @fields.depends('company')
-    def on_change_with_currency_digits(self, name=None):
-        if self.company:
-            return self.company.currency.digits
-        return 2
+    def on_change_with_currency(self, name=None):
+        return self.company.currency if self.company else None
 
     @classmethod
     def copy(cls, taxes, default=None):
@@ -1033,7 +1084,7 @@ class Tax(sequence_ordered(), ModelSQL, ModelView, DeactivableMixin):
         return sorted(taxes, key=key, reverse=reverse)
 
     @classmethod
-    def compute(cls, taxes, price_unit, quantity, date=None):
+    def compute(cls, taxes, price_unit, quantity, date):
         '''
         Compute taxes for price_unit and quantity at the date.
         Return list of dict for each taxes and their childs with:
@@ -1041,10 +1092,6 @@ class Tax(sequence_ordered(), ModelSQL, ModelView, DeactivableMixin):
             amount
             tax
         '''
-        pool = Pool()
-        Date = pool.get('ir.date')
-        if date is None:
-            date = Date.today()
         taxes = cls.sort_taxes(taxes)
         res = cls._unit_compute(taxes, price_unit, date)
         quantity = Decimal(str(quantity or 0))
@@ -1054,14 +1101,10 @@ class Tax(sequence_ordered(), ModelSQL, ModelView, DeactivableMixin):
         return res
 
     @classmethod
-    def reverse_compute(cls, price_unit, taxes, date=None):
+    def reverse_compute(cls, price_unit, taxes, date):
         '''
         Reverse compute the price_unit for taxes at the date.
         '''
-        pool = Pool()
-        Date = pool.get('ir.date')
-        if date is None:
-            date = Date.today()
         taxes = cls.sort_taxes(taxes)
         return cls._reverse_unit_compute(price_unit, taxes, date)
 
@@ -1180,7 +1223,8 @@ class TaxableMixin(object):
         "Date to use when computing the tax"
         pool = Pool()
         Date = pool.get('ir.date')
-        return Date.today()
+        with Transaction().set_context(company=self.company.id):
+            return Date.today()
 
     def _get_tax_context(self):
         return {}
@@ -1206,10 +1250,25 @@ class TaxableMixin(object):
     def _round_taxes(self, taxes):
         if not self.currency:
             return
-        for taxline in taxes.values():
-            taxline['amount'] = self.currency.round(taxline['amount'])
 
-    @fields.depends(methods=['_get_tax_context', '_round_taxes'])
+        remainder = 0
+        for taxline in taxes.values():
+            rounded_amount = self.currency.round(taxline['amount'])
+            remainder += rounded_amount - taxline['amount']
+            taxline['amount'] = rounded_amount
+
+        # We need to compensate the rounding we did
+        remainder = self.currency.round(remainder, opposite=True)
+        if abs(remainder) >= self.currency.rounding:
+            offset_amount = self.currency.rounding.copy_sign(remainder)
+
+            for tax in cycle(taxes.values()):
+                tax['amount'] -= offset_amount
+                remainder -= offset_amount
+                if abs(remainder) < self.currency.rounding:
+                    break
+
+    @fields.depends('company', methods=['_get_tax_context', '_round_taxes'])
     def _get_taxes(self):
         pool = Pool()
         Tax = pool.get('account.tax')
@@ -1222,8 +1281,10 @@ class TaxableMixin(object):
             taxable_lines = [_TaxableLine(*params)
                 for params in self.taxable_lines]
             for line in taxable_lines:
+                assert all(t.company == self.company for t in line.taxes)
                 l_taxes = Tax.compute(Tax.browse(line.taxes), line.unit_price,
                     line.quantity, line.tax_date or self.tax_date)
+                current_taxes = {}
                 for tax in l_taxes:
                     taxline = self._compute_tax_line(**tax)
                     # Base must always be rounded per line as there will be one
@@ -1235,8 +1296,9 @@ class TaxableMixin(object):
                     else:
                         taxes[taxline]['base'] += taxline['base']
                         taxes[taxline]['amount'] += taxline['amount']
+                    current_taxes[taxline] = taxes[taxline]
                 if tax_rounding == 'line':
-                    self._round_taxes(taxes)
+                    self._round_taxes(current_taxes)
         if tax_rounding == 'document':
             self._round_taxes(taxes)
         return taxes
@@ -1245,22 +1307,22 @@ class TaxableMixin(object):
 class TaxLine(ModelSQL, ModelView):
     'Tax Line'
     __name__ = 'account.tax.line'
-    currency_digits = fields.Function(fields.Integer('Currency Digits'),
-        'on_change_with_currency_digits')
-    amount = fields.Numeric('Amount', digits=(16, Eval('currency_digits', 2)),
-        required=True, depends=['currency_digits'])
+    currency = fields.Function(fields.Many2One(
+            'currency.currency', "Currency"),
+        'on_change_with_currency')
+    amount = Monetary(
+        "Amount", currency='currency', digits='currency', required=True)
     type = fields.Selection([
             ('tax', "Tax"),
             ('base', "Base"),
             ], "Type", required=True)
-    tax = fields.Many2One('account.tax', 'Tax', select=True,
-        ondelete='RESTRICT', required=True,
+    tax = fields.Many2One(
+        'account.tax', "Tax", ondelete='RESTRICT', required=True,
         domain=[
             ('company', '=', Eval('company', -1)),
-            ],
-        depends=['company'])
-    move_line = fields.Many2One('account.move.line', 'Move Line',
-            required=True, select=True, ondelete='CASCADE')
+            ])
+    move_line = fields.Many2One(
+        'account.move.line', "Move Line", required=True, ondelete='CASCADE')
     company = fields.Function(fields.Many2One('company.company', 'Company'),
         'on_change_with_company')
 
@@ -1310,16 +1372,14 @@ class TaxLine(ModelSQL, ModelView):
                         & (table.code.in_(
                                 [invoice_base_code, credit_note_base_code]))))
 
-    @fields.depends('move_line', '_parent_move_line.currency_digits')
-    def on_change_with_currency_digits(self, name=None):
-        if self.move_line:
-            return self.move_line.currency_digits
-        return 2
+    @fields.depends('move_line', '_parent_move_line.currency')
+    def on_change_with_currency(self, name=None):
+        return self.move_line.currency if self.move_line else None
 
     @fields.depends('_parent_move_line.account', 'move_line')
     def on_change_with_company(self, name=None):
         if self.move_line and self.move_line.account:
-            return self.move_line.account.company.id
+            return self.move_line.account.company
 
     def get_rec_name(self, name):
         name = super(TaxLine, self).get_rec_name(name)
@@ -1429,16 +1489,14 @@ class TaxRule(ModelSQL, ModelView):
         }
     name = fields.Char('Name', required=True, states=_states)
     kind = fields.Selection(KINDS, 'Kind', required=True, states=_states)
-    company = fields.Many2One(
-        'company.company', "Company", required=True, select=True)
+    company = fields.Many2One('company.company', "Company", required=True,)
     lines = fields.One2Many('account.tax.rule.line', 'rule', 'Lines')
     template = fields.Many2One('account.tax.rule.template', 'Template')
     template_override = fields.Boolean("Override Template",
         help="Check to override template definition",
         states={
             'invisible': ~Bool(Eval('template', -1)),
-            },
-        depends=['template'])
+            })
     del _states
 
     @staticmethod
@@ -1528,7 +1586,7 @@ class TaxRuleLineTemplate(sequence_ordered(), ModelSQL, ModelView):
             ],
         help=('If the original tax template is filled, the rule will be '
             'applied only for this tax template.'),
-        depends=['group'],
+        depends={'rule'},
         ondelete='RESTRICT')
     keep_origin = fields.Boolean("Keep Origin",
         help="Check to append the original tax to substituted tax.")
@@ -1547,7 +1605,7 @@ class TaxRuleLineTemplate(sequence_ordered(), ModelSQL, ModelView):
                         ('group.kind', 'in', ['sale', 'purchase', 'both']))),
                 ],
             ],
-        depends=['group'],
+        depends={'rule'},
         ondelete='RESTRICT')
 
     @classmethod
@@ -1626,8 +1684,9 @@ class TaxRuleLine(sequence_ordered(), ModelSQL, ModelView, MatchMixin):
         'readonly': (Bool(Eval('template', -1))
             & ~Eval('template_override', False)),
         }
-    rule = fields.Many2One('account.tax.rule', 'Rule', required=True,
-            select=True, ondelete='CASCADE', states=_states)
+    rule = fields.Many2One(
+        'account.tax.rule', "Rule",
+        required=True, ondelete='CASCADE', states=_states)
     start_date = fields.Date("Starting Date")
     end_date = fields.Date("Ending Date")
     group = fields.Many2One('account.tax.group', 'Tax Group',
@@ -1649,7 +1708,7 @@ class TaxRuleLine(sequence_ordered(), ModelSQL, ModelView, MatchMixin):
             ],
         help=('If the original tax is filled, the rule will be applied '
             'only for this tax.'),
-        depends=['group'],
+        depends={'rule'},
         ondelete='RESTRICT', states=_states)
     keep_origin = fields.Boolean("Keep Origin", states=_states,
         help="Check to append the original tax to substituted tax.")
@@ -1668,7 +1727,7 @@ class TaxRuleLine(sequence_ordered(), ModelSQL, ModelView, MatchMixin):
                         ('group.kind', 'in', ['sale', 'purchase', 'both']))),
                 ],
             ],
-        depends=['group'],
+        depends={'rule'},
         ondelete='RESTRICT', states=_states)
     template = fields.Many2One(
         'account.tax.rule.line.template', 'Template', ondelete='CASCADE')
@@ -1676,8 +1735,7 @@ class TaxRuleLine(sequence_ordered(), ModelSQL, ModelView, MatchMixin):
         help="Check to override template definition",
         states={
             'invisible': ~Bool(Eval('template', -1)),
-            },
-        depends=['template'])
+            })
     del _states
 
     @classmethod
@@ -1848,6 +1906,14 @@ class TestTaxView(ModelView, TaxableMixin):
             company = Company(company_id)
             return company.currency.id
 
+    @property
+    def company(self):
+        pool = Pool()
+        Company = pool.get('company.company')
+        company_id = Transaction().context.get('company')
+        if company_id:
+            return Company(company_id)
+
     @property
     def taxable_lines(self):
         return [(self.taxes, self.unit_price, self.quantity, self.tax_date)]
@@ -1862,8 +1928,7 @@ class TestTaxView(ModelView, TaxableMixin):
             for taxline in self._get_taxes():
                 del taxline['manual']
                 result.append(Result(**taxline))
-        self.result = result
-        return self._changed_values.get('result', [])
+        return result
 
 
 class TestTaxViewResult(ModelView):
